/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package net.java.spnego;

import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.Key;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.crypto.Cipher;
import javax.crypto.Mac;
import javax.crypto.SecretKey;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.kerberos.KerberosKey;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;
import javax.security.auth.message.AuthException;
import javax.security.auth.message.MessageInfo;
import javax.security.auth.message.MessagePolicy;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import net.java.spnego.pac.Pac;
import net.java.spnego.pac.PacLogonInfo;
import net.java.spnego.pac.PacSid;
import org.apache.commons.codec.binary.Base64;
import org.ietf.jgss.GSSContext;
import sun.security.util.DerInputStream;
import sun.security.util.DerValue;
import sun.security.util.ObjectIdentifier;

/**
 * <p> PAC SPENEGO server authentication module. The module parses the 
 * SPNEGO/Kerberos token in order to get the PAC info (SID for groups) the user
 * belongs to. This class is based in the Spring implementation.</p>
 * 
 * <p>The token is parsed using sun.security.util.Der* classes for ASN.1 parsing.
 * A Kerberos subject from the loginContext is used to decrypt the encData
 * in the token (the subject has the server keys). All the implementation is based
 * in <a href="http://jaaslounge.sourceforge.net/">jaaslounge</a> implementation, 
 * to the extent that all the PAC classes are copied (with minimal changes).</p>
 * 
 * <p>References:</p>
 * 
 * <ol>
 * <li><a href="http://git.springsource.org/~grantcermak/spring-security/activedirectory-se-security/blobs/b95de8e107937ed6bab58528561e337c910f769a/spring-security-kerberos/spring-security-kerberos-core/src/main/java/org/springframework/security/extensions/kerberos/ActiveDirectorySecurityIntegration.java">ActiveDirectorySecurityIntegration</a>: 
 * Spring class that does exactly the same than here.</li>
 * <li><a href="http://jaaslounge.svn.sourceforge.net/viewvc/jaaslounge/trunk/decoding/">org.jaaslounge.decoding</a>:
 * Package that performs the SPNEGO/Kerberos token parsing. These classes uses the
 * BouncyCastle ASN.1 parsing implementation. Here sun.security.util (DerInputStream, DerValue and so on) 
 * package is used.</li>
 * <ol>
 * 
 * @author ricky
 */
public class PACSpnegoServerAuth extends SpnegoServerAuthModule {

    //
    // CONSTANTS
    //
    
    /**
     * logger for the class
     */
    private static final Logger LOG = Logger.getLogger(PACSpnegoServerAuth.class.getName());
    
    /**
     * Attribute in the session in which the SIDs will be added
     */
    private static final String SIDS_SESSION_ATTR = "__SPNEGO_PAC_SIDS__";
    
    /**
     * Default login module for kerberos
     */
    private static final String KRB5_LOGIN_MODULE = "com.sun.security.jgss.accept";
    
    /**
     * SPNEGO OID mechanism to detect in the token
     */
    private static final String SPNEGO_MECHANISM = "1.3.6.1.5.5.2";
    
    /**
     * Kerberos OID to detect in the tocken
     */
    private static final String KERBEROS_OID = "1.2.840.113554.1.2.2";
    
    /**
     * Mask to detect the tag in DER encoding
     */
    private static final byte TAG_NO_MASK = (byte) 0x1F;
    
    /**
     * Authentication data relevant type
     */
    private static final int AUTH_DATA_RELEVANT = 1;
    
    /**
     * Authentication data PAC type
     */
    private static final int AUTH_DATA_PAC = 128;
    
    /**
     * DES encode type
     */
    private static final int DES_ENC_TYPE = 3;
    
    /**
     * RC4 encode type
     */
    private static final int RC4_ENC_TYPE = 23;
    
    /**
     * the checksum size
     */
    private static final int CHECKSUM_SIZE = 16;
    
    /**
     * RC4 algorithm
     */
    private static final String RC4_ALGORITHM = "ARCFOUR";
    
    /**
     * founder size
     */
    private static final int CONFOUNDER_SIZE = 8;
    
    /**
     * HMAC algorithm
     */
    private static final String HMAC_ALGORITHM = "HmacMD5";
    
    //
    // PROPERTIES
    //
    
    /**
     * The server credentials used to decode the encrypted part of the Kerberos token
     * where the PAC is
     */
    private Set<KerberosKey> serverCredentials = null;
    
    /**
     * The login context used for login (KRB5_LOGIN_MODULE)
     */
    private LoginContext loginContext = null;

    /**
     * It just calls the super and then get the login context (KRB5_LOGIN_MODULE)
     * and logins and obtains the server keys for later decrypting.
     * @param requestPolicy The request policy this module must enforce, or null.
     * @param responsePolicy The response policy this module must enforce, or null.
     * @param handler CallbackHandler used to request information.
     * @param options A Map of module-specific configuration properties.
     * @throws AuthException Some error initializing the module
     */
    @Override
    public void initialize(MessagePolicy requestPolicy,
            MessagePolicy responsePolicy,
            CallbackHandler handler,
            Map options)
            throws AuthException {
        super.initialize(requestPolicy, responsePolicy, handler, options);
        try {
            loginContext = new LoginContext(KRB5_LOGIN_MODULE);
            loginContext.login();
            serverCredentials = loginContext.getSubject().getPrivateCredentials(KerberosKey.class);
        } catch (LoginException e) {
            LOG.log(Level.WARNING, "Error login in the Kerberos login module!", e);
            AuthException ae = new AuthException(e.getLocalizedMessage());
            ae.initCause(e);
            throw ae;
        }
    }

    /**
     * Method that parses the SPNEGO (gss) token in order to get the
     * kerberos token.
     * @param gssToken The SPNEGO/GSS token to parse
     * @return The kerberos mechanism token associated with the SPNEGO token
     * @throws IOException Some error parsing
     */
    public byte[] getMechanismToken(byte[] gssToken) throws IOException {
        // read the application
        DerInputStream stream = new DerInputStream(gssToken);
        DerValue value = stream.getDerValue();
        if (!value.isConstructed() || !value.isApplication()) {
            throw new IOException("Invalid SPNEGO token => Application is not OK!");
        }
        stream = value.getData();
        // read the spnego token OID
        value = stream.getDerValue();
        ObjectIdentifier oid = value.getOID();
        if (!SPNEGO_MECHANISM.equals(oid.toString())) {
            throw new IOException("Invalid SPNEGO token => Mech OID is not SPNEGO!");
        }
        // read the sequence to get the mechanism token
        value = stream.getDerValue();
        stream = value.getData();
        value = stream.getDerValue();
        stream = value.toDerInputStream();
        while (stream.available() > 0) {
            value = stream.getDerValue();
            if ((value.getTag() & TAG_NO_MASK) == 0x02) {
                // 0x02 is the mechanism token application
                stream = value.getData();
                value = stream.getDerValue();
                //System.err.println("RICKY: " + String.format("%#4x", new Byte(value.getTag()).intValue() & 0xFF));
                return value.getOctetString();
            }
        }
        throw new IOException("Invalid SPNEGO token => No mechanism token!");
    }

    /**
     * This method parses the kerberos token and returns the Kerberos AP request.
     * @param mechanismToken The kerberos mechanism token to parse
     * @return The Kerberos AP request token
     * @throws IOException Some error parsing
     */
    public byte[] getKerberosApRequest(byte[] mechanismToken) throws IOException {
        DerInputStream stream = new DerInputStream(mechanismToken);
        DerValue value = stream.getDerValue();
        if (!value.isConstructed() || !value.isApplication()) {
            throw new IOException("Invalid mechanism token => Application is not OK!");
        }
        stream = value.getData();
        // read the kerberos OID
        value = stream.getDerValue();
        ObjectIdentifier oid = value.getOID();
        if (!KERBEROS_OID.equals(oid.toString())) {
            throw new IOException("Invalid mechanism token => Invalid Kerberos OID!");
        }
        // check version
        byte[] readArray = new byte[2];
        stream.getBytes(readArray);
        int readLow = readArray[0] & 0xff;
        int readHigh = readArray[1] & 0xff;
        int read = (readHigh << 8) + readLow;
        if (read != 0x01) {
            throw new IOException("Invalid mechanism token => Mech OID is not SPNEGO!");
        }
        // read the KerberosApRequest
        value = stream.getDerValue();
        if (!value.isConstructed() || !value.isApplication()) {
            throw new IOException("Invalid mechanism token => Application is not OK!");
        }
        return value.getDataBytes();
    }
    
    /**
     * Method that parses the Kerberos AP request in order to get the kerberos ticket.
     * @param kerberosApRequest The kerberos AP request token
     * @return The kerberos ticket
     * @throws IOException Some error parsing
     */
    public byte[] getKerberosTicket(byte[] kerberosApRequest) throws IOException {
        DerInputStream stream = new DerInputStream(kerberosApRequest);
        DerValue value = stream.getDerValue();
        // it's a sequence
        stream = value.toDerInputStream();
        while (stream.available() > 0) {
            value = stream.getDerValue();
            //System.err.println("RICKY: " + String.format("%#4x", new Byte(value.getTag()).intValue() & 0xFF));
            if ((value.getTag() & TAG_NO_MASK) == 0x03) {
                // 0x03 is the kerberos ticket application
                stream = value.getData();
                value = stream.getDerValue();
                if (!value.isConstructed() || !value.isApplication()) {
                    throw new IOException("Invalid KerberosApRequest token => Application is not OK!");
                }
                return value.getDataBytes();
            }
        }
        throw new IOException("Invalid KerberosApRequest token => !");
    }
    
    /**
     * Return a inetger as a byte array
     * @param integer The integer to conver into a byte array
     * @return The byte array that represents the integer
     */
    private static byte[] asBytes(int integer) {
        byte[] bytes = new byte[]{(byte)integer, (byte)(integer >>> 8), (byte)(integer >>> 16),
                (byte)(integer >>> 24)};

        return bytes;
    }
    
    /**
     * 
     * @param data
     * @param key
     * @return
     * @throws GeneralSecurityException 
     */
    private static byte[] getHmac(byte[] data, byte[] key) throws GeneralSecurityException {
        Key macKey = new SecretKeySpec(key.clone(), HMAC_ALGORITHM);
        Mac mac = Mac.getInstance(HMAC_ALGORITHM);
        mac.init(macKey);

        return mac.doFinal(data);
    }
    
    /**
     * Method that decrypts the encrypted data section of the token.
     * @param data The encrypted data section read from the token
     * @param key The server key used to decrypt
     * @param type The type of the key
     * @return The decrypted data
     * @throws GeneralSecurityException Some error performing the operation
     */
    public static byte[] decrypt(byte[] data, Key key, int type) throws GeneralSecurityException {
        Cipher cipher = null;
        byte[] decrypt = null;

        switch (type) {
        case DES_ENC_TYPE:
            try {
                cipher = Cipher.getInstance("DES/CBC/NoPadding");
            } catch(GeneralSecurityException e) {
                throw new GeneralSecurityException("Checksum failed while decrypting.");
            }
            byte[] ivec = new byte[8];
            IvParameterSpec params = new IvParameterSpec(ivec);
            
            SecretKeySpec skSpec = new SecretKeySpec(key.getEncoded(), "DES");
            SecretKey sk = (SecretKey)skSpec;

            cipher.init(Cipher.DECRYPT_MODE, sk, params);
            
            byte[] result;
            result = cipher.doFinal(data);
            
            decrypt = new byte[result.length];
            System.arraycopy(result, 0, decrypt, 0, result.length);
            
            int tempSize = decrypt.length - 24;
            
            byte[] output = new byte[tempSize];
            System.arraycopy(decrypt, 24, output, 0, tempSize);
            
            decrypt = output;
            break;
        case RC4_ENC_TYPE:
            byte[] code = asBytes(Cipher.DECRYPT_MODE);
            byte[] codeHmac = getHmac(code, key.getEncoded());

            byte[] dataChecksum = new byte[CHECKSUM_SIZE];
            System.arraycopy(data, 0, dataChecksum, 0, CHECKSUM_SIZE);

            byte[] dataHmac = getHmac(dataChecksum, codeHmac);
            SecretKeySpec dataKey = new SecretKeySpec(dataHmac, RC4_ALGORITHM);

            cipher = Cipher.getInstance(RC4_ALGORITHM);
            cipher.init(Cipher.DECRYPT_MODE, dataKey);

            int plainDataLength = data.length - CHECKSUM_SIZE;
            byte[] plainData = cipher.doFinal(data, CHECKSUM_SIZE,
                    plainDataLength);

            byte[] plainDataChecksum = getHmac(plainData, codeHmac);
            if(plainDataChecksum.length >= CHECKSUM_SIZE) {
                for(int i = 0; i < CHECKSUM_SIZE; i++) {
                    if(plainDataChecksum[i] != data[i]) {
                        throw new GeneralSecurityException("Checksum failed while decrypting.");
                    }
                }
            }
            int decryptLength = plainData.length - CONFOUNDER_SIZE;
            decrypt = new byte[decryptLength];
            System.arraycopy(plainData, CONFOUNDER_SIZE, decrypt, 0,
                    decryptLength);
            break;
        default:
            throw new GeneralSecurityException("Unsupported encryption type.");
        }
        return decrypt;
    }
    
    /**
     * Method that parses the kerberos ticket to find the encrypted data. The
     * data is decrypted before returning (so the byte array returned is the
     * decrypted data).
     * @param kerberosTicket The kerberos ticket to parse
     * @param keys The server keys in order to decrypt the data
     * @return The encrypted data section but decrypted
     * @throws IOException Some error parsing
     * @throws GeneralSecurityException Some error decrypting
     */
    public byte[] getEncodedData(byte[] kerberosTicket, KerberosKey[] keys) throws IOException, GeneralSecurityException {
        DerInputStream stream = new DerInputStream(kerberosTicket);
        DerValue value = stream.getDerValue();
        // it's a sequence
        stream = value.toDerInputStream();
        while (stream.available() > 0) {
            value = stream.getDerValue();
            //System.err.println("RICKY: " + String.format("%#4x", new Byte(value.getTag()).intValue() & 0xFF));
            if ((value.getTag() & TAG_NO_MASK) == 0x03) {
                // 0x03 is the kerberos enc data application
                value = value.getData().getDerValue();
                // there's again a der sequence inside => encType and encOctects
                stream = value.toDerInputStream();
                int encType = -1;
                byte[] encData = null;
                while (stream.available() > 0) {
                    value = stream.getDerValue();
                    //System.err.println("RICKY en 3: " + String.format("%#4x", new Byte(value.getTag()).intValue() & 0xFF));
                    if ((value.getTag() & TAG_NO_MASK) == 0x00) {
                        // 0x00 is the encType
                        value = value.getData().getDerValue();
                        encType = value.getInteger();
                        //System.err.println("RICKY encType: " + encType);
                    } else if ((value.getTag() & TAG_NO_MASK) == 0x02) {
                        // 0x02 is the encData
                        value = value.getData().getDerValue();
                        encData = value.getOctetString();
                        //System.err.println("RICKY encData: " + encData);
                    }
                }
                if (encType == -1 || encData == null) {
                    throw new IOException("Invalid kerberos ticket => No suitable encType or encData!");
                }
                // locate the server key for this encType
                KerberosKey serverKey = null;
                for (KerberosKey key : keys) {
                    if (key.getKeyType() == encType) {
                        serverKey = key;
                    }
                }
                if (serverKey == null) {
                    throw new IOException("Invalid kerberos ticket => No suitable server key!");
                }
                // decrypt the enc data
                return decrypt(encData, serverKey, serverKey.getKeyType());
            }
        }
        throw new IOException("Invalid kerberos ticket => No enc data in the ticket!");
    }
    
    /**
     * Method that parses the PAC section in order to add all PAC information.
     * This method is recursive.
     * @param value The DerValue with PAC information
     * @param pacs The list of pacs (new ones will be added here)
     * @throws IOException Some error parsing
     */
    private void getPacs(DerValue value, List<Pac> pacs) throws IOException {
        DerInputStream authStream = value.toDerInputStream();
        int authType = -1;
        byte[] authData = null;
        while (authStream.available() > 0) {
            value = authStream.getDerValue();
            //System.err.println("RICKY getPacs: " + String.format("%#4x", new Byte(value.getTag()).intValue() & 0xFF));
            if ((value.getTag() & TAG_NO_MASK) == 0x00) {
                // 0x00 is the authType
                value = value.getData().getDerValue();
                authType = value.getInteger();
                //System.err.println("RICKY getPacs authType=" + authType);
            } else if ((value.getTag() & TAG_NO_MASK) == 0x01) {
                // 0x01 is the authData
                value = value.getData().getDerValue();
                authData = value.getOctetString();
                //System.err.println("RICKY getPacs authData=" + authData);
            }
        }
        if (authType == AUTH_DATA_RELEVANT && authData != null) {
            value = new DerValue(authData);
            getPacs(value.getData().getDerValue(), pacs);
        } else if (authType == AUTH_DATA_PAC && authData != null) {
            //System.err.println("RICKY getPacs adding...");
            pacs.add(new Pac(authData));
        }
    }
    
    /**
     * Method that parses the encrypted data (sgould be already decrypted) to
     * get all the PAC information.
     * @param encData The encrypted data but decrypted
     * @return The list of PACs the data contains
     * @throws IOException Some error parsing
     */
    private List<Pac> getPacsFromEncData(byte[] encData) throws IOException {
        List<Pac> pacs = new ArrayList<Pac>();
        DerInputStream stream = new DerInputStream(encData);
        DerValue value = stream.getDerValue();
        if (!value.isConstructed() || !value.isApplication()) {
            throw new IOException("Invalid EncData token => Application is not OK!");
        }
        value = value.getData().getDerValue();
        // it's again a DER sequence
        stream = value.toDerInputStream();
        while (stream.available() > 0) {
            value = stream.getDerValue();
            //System.err.println("RICKY: " + String.format("%#4x", new Byte(value.getTag()).intValue() & 0xFF));
            if ((value.getTag() & TAG_NO_MASK) == 0x0A) {
                // 0x0A is the authorizations data and again a DER sequence
                value = value.getData().getDerValue();
                stream = value.toDerInputStream();
                while (stream.available() > 0) {
                    // each auth is again a DER sequence
                    value = stream.getDerValue();
                    getPacs(value, pacs);
                }
                return pacs;
            }
        }
        throw new IOException("Invalid EncData token => No authorizations!");
    }

    /**
     * This method is intended to add all the group information for a 
     * SPNEGO auth module. In this case the GSS token is retrieved again and all
     * the PAC section is obtained to add all the SIDs as groups.
     * @param messageInfo A contextual object that encapsulates the
     *                    client request and server response objects, and that may be
     *                    used to save state across a sequence of calls made to the
     *                    methods of this interface for the purpose of completing a
     *                    secure message exchange.
     * @param gssContext The GSSContext used to accept the security context
     * @return The list of groups (SIDs) the contain contains
     * @throws AuthException Some error in the parsing of the token
     */
    @Override
    public String[] getGroupsForCaller(MessageInfo messageInfo, GSSContext gssContext
            ) throws AuthException {
        List<String> sids = null;
        HttpServletRequest request = (HttpServletRequest) messageInfo.getRequestMessage();
        HttpSession sess = request.getSession(false);
        if (sess != null && sess.getAttribute(SIDS_SESSION_ATTR) != null) {
            LOG.info("Getting SIDs from the session...");
            sids = (List<String>) sess.getAttribute(SIDS_SESSION_ATTR);
        } else {
            try {
                // add the parsing
                LOG.info("First time! Let's parse the kerberos token...");
                sids = new ArrayList<String>();
                // create the parser SPNEGO token => it's sure all is ok (do not needs checks)
                String header = request.getHeader(SpnegoServerAuthModule.AUTHORIZATION_HEADER);
                header = header.substring(SpnegoServerAuthModule.NEGOTIATE.length() + 1);
                byte[] requestToken = Base64.decodeBase64(header.getBytes());
                // parse the SPNEGO token to get the mechanism (kerberos) one
                byte[] mechanismToken = getMechanismToken(requestToken);
                KerberosKey[] keys = serverCredentials.toArray(new KerberosKey[serverCredentials.size()]);
                // parse kerberos token to get the PAC login Info
                byte[] kerberosApRequest = getKerberosApRequest(mechanismToken);
                byte[] kerberosTicket = getKerberosTicket(kerberosApRequest);
                byte[] encData = getEncodedData(kerberosTicket, keys);
                List<Pac> pacs = getPacsFromEncData(encData);
                // iterate the pacs to get the groups
                for (Pac pac: pacs) {
                    // get the PAC info
                    PacLogonInfo logonInfo = pac.getLogonInfo();
                    if (logonInfo != null) {
                        if (logonInfo.getGroupSid() != null) {
                            String sid = logonInfo.getGroupSid().toString();
                            LOG.log(Level.INFO, "Adding primary group SID: {0}", sid);
                            sids.add(sid);
                        }
                        if (logonInfo.getGroupSids() != null) {
                            for (PacSid pacSid : logonInfo.getGroupSids()) {
                                String sid = pacSid.toString();
                                LOG.log(Level.INFO, "Adding secondary group SID: {0}", sid);
                                sids.add(sid);
                            }
                        }
                        if (logonInfo.getExtraSids() != null) {
                            for (PacSid pacSid : logonInfo.getExtraSids()) {
                                String sid = pacSid.toString();
                                LOG.log(Level.INFO, "Adding extra group SID: {0}", sid);
                                sids.add(sid);
                            }
                        }
                        if (logonInfo.getResourceGroupSids() != null) {
                            for (PacSid pacSid : logonInfo.getResourceGroupSids()) {
                                String sid = pacSid.toString();
                                LOG.log(Level.INFO, "Adding resource group SID: {0}", sid);
                                sids.add(sid);
                            }
                        }
                    }
                }
                // add the SIDs to the session
                request.getSession(true).setAttribute(SIDS_SESSION_ATTR, sids);
            } catch (Exception e) {
                LOG.log(Level.WARNING, "Error decoding PAC!!!", e);
                AuthException ae = new AuthException(e.getLocalizedMessage());
                ae.initCause(e);
                throw ae;
            }
        }
        return sids.toArray(new String[0]);
    }

    /**
     * Logout from the login context
     * @throws Throwable 
     */
    @Override
    protected void finalize() throws Throwable {
        super.finalize();
        if (null != this.loginContext) {
            try {
                this.loginContext.logout();
            } catch (LoginException le) {
                LOG.log(Level.WARNING, "Logout failed.", le);
            }
        }
    }
}
