////////////////////////////////////////////////////////////////////////////////
//
// The contents of this file are subject to the terms of the Common Development
// and Distribution License(the License).  You may not use this file except in
// compliance with the License.
//
// You may obtain a copy of the License from:
// http://www.opensource.org/licenses/cddl1.txt
// See the License for the specific language governing
// permissions and limitations under the License.
//
// Copyright (c) 2007, Java Spnego Development Team.
// (http://spnego.dev.java.net/servlets/ProjectMemberList) All Rights Reserved.
//
////////////////////////////////////////////////////////////////////////////////

package net.java.spnego;

//GSSUtil is in the JDK, not Glassfish.

import com.sun.security.jgss.GSSUtil;
import org.apache.commons.codec.binary.Base64;
import org.ietf.jgss.GSSContext;
import org.ietf.jgss.GSSCredential;
import org.ietf.jgss.GSSException;
import org.ietf.jgss.GSSManager;
import org.ietf.jgss.GSSName;
import org.ietf.jgss.Oid;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.message.AuthException;
import javax.security.auth.message.AuthStatus;
import javax.security.auth.message.MessageInfo;
import javax.security.auth.message.MessagePolicy;
import javax.security.auth.message.callback.CallerPrincipalCallback;
import javax.security.auth.message.callback.GroupPrincipalCallback;
import javax.security.auth.message.module.ServerAuthModule;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.security.Principal;
import java.util.Enumeration;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This JMAC Module supports the SPNEGO mechanism.
 * <p/>
 * Kerberos is implemented but not NTLM, yet.
 * <p/>
 * This module must be configured in Glassfish to take effect.
 *
 * @author Greg Luck, Ron Monzillo
 * @version $Id: TestLoginModule.java 0 Feb 28, 2007 10:11:43 PM dwhitla $
 * @see javax.security.auth.message.MessageInfo
 * @see javax.security.auth.Subject
 */
public abstract class SpnegoServerAuthModule implements ServerAuthModule {

    public static final String AUTH_TYPE_INFO_KEY = "javax.servlet.http.authType";

    private static final Logger LOG = Logger.getLogger(SpnegoServerAuthModule.class.getName());

    private static final String DEBUG_OPTIONS_KEY = "debug";

    private static final String POLICY_CONTEXT_OPTIONS_KEY = "javax.security.jacc.PolicyContext";

    private static final String IS_MANDATORY_INFO_KEY = "javax.security.auth.message.MessagePolicy.isMandatory";

    protected static final String AUTHORIZATION_HEADER = "authorization";

    private static final String AUTHENTICATION_HEADER = "WWW-Authenticate";

    protected static final String NEGOTIATE = "Negotiate";

    private static final String NTLM_INITIAL_TOKEN = "NTLMSSP";

    private static Class[] supportedMessageTypes = new Class[]{
            javax.servlet.http.HttpServletRequest.class, javax.servlet.http.HttpServletResponse.class};

    private MessagePolicy requestPolicy;

    private MessagePolicy responsePolicy;

    private CallbackHandler handler;

    private Map options;

    private boolean debug;

    private Level debugLevel;

    private String policyContextID;

    private boolean mandatory;

    private GSSManager gssManager;

    /**
     * Initialize this module with request and response message policies
     * to enforce, a CallbackHandler, and any module-specific configuration
     * properties.
     * <p/>
     * <p> The request policy and the response policy must not both be null.
     *
     * @param requestPolicy  The request policy this module must enforce,
     *                       or null.
     * @param responsePolicy The response policy this module must enforce,
     *                       or null.
     * @param handler        CallbackHandler used to request information.
     * @param options        A Map of module-specific configuration properties.
     *                       //todo checkstyle is barfing on this
     *                       throws AuthException If module initialization fails, including for
     *                       the case where the options argument contains elements that are not
     *                       supported by the module.
     */
    public void initialize(MessagePolicy requestPolicy,
                           MessagePolicy responsePolicy,
                           CallbackHandler handler,
                           Map options)
            throws AuthException {

        this.requestPolicy = requestPolicy;
        this.responsePolicy = responsePolicy;

        this.mandatory = requestPolicy.isMandatory();

        this.handler = handler;
        this.options = options;

        if (options != null) {
            debug = options.containsKey(DEBUG_OPTIONS_KEY);
            policyContextID = (String)
                    options.get(POLICY_CONTEXT_OPTIONS_KEY);
        } else {
            debug = false;
            policyContextID = null;
        }

        debugLevel = (LOG.isLoggable(Level.FINE) && !debug) ? Level.FINE : Level.INFO;

        gssManager = GSSManager.getInstance();
    }

    /**
     * Get the one or more Class objects representing the message types
     * supported by the module.
     *
     * @return An array of Class objects, with at least one element
     *         defining a message type supported by the module.
     */
    public Class[] getSupportedMessageTypes() {
        return supportedMessageTypes;
    }

    /**
     * Authenticate a received service request.
     * <p/>
     * This method is called to transform the mechanism-specific request
     * message acquired by calling getRequestMessage (on messageInfo)
     * into the validated application message to be returned to the message
     * processing runtime.
     * If the received message is a (mechanism-specific) meta-message,
     * the method implementation must attempt to transform the meta-message
     * into a corresponding mechanism-specific response message, or to the
     * validated application request message.
     * The runtime will bind a validated application message into the
     * the corresponding service invocation.
     * <p> This method conveys the outcome of its message processing either
     * by returning an AuthStatus value or by throwing an AuthException.
     * <p/>
     * From a performance point of view this method will be called twice for each
     * resource with a secuirty constraint on it. Resources with no security constraint
     * do not result in a call to this method.
     *
     * @param messageInfo    A contextual object that encapsulates the
     *                       client request and server response objects, and that may be
     *                       used to save state across a sequence of calls made to the
     *                       methods of this interface for the purpose of completing a
     *                       secure message exchange.
     * @param clientSubject  A Subject that represents the source of the
     *                       service
     *                       request.  It is used by the method implementation to store
     *                       Principals and credentials validated in the request.
     * @param serviceSubject A Subject that represents the recipient of the
     *                       service request, or null.  It may be used by the method
     *                       implementation as the source of Principals or credentials to
     *                       be used to validate the request. If the Subject is not null,
     *                       the method implementation may add additional Principals or
     *                       credentials (pertaining to the recipient of the service
     *                       request) to the Subject.
     * @return An AuthStatus object representing the completion status of
     *         the processing performed by the method.
     *         The AuthStatus values that may be returned by this method
     *         are defined as follows:
     *         <p/>
     *         <ul>
     *         <li> AuthStatus.SUCCESS when the application request message
     *         was successfully validated. The validated request message is
     *         available by calling getRequestMessage on messageInfo.
     *         <p/>
     *         <li> AuthStatus.SEND_SUCCESS to indicate that validation/processing
     *         of the request message successfully produced the secured application
     *         response message (in messageInfo). The secured response message is
     *         available by calling getResponseMessage on messageInfo.
     *         <p/>
     *         <li> AuthStatus.SEND_CONTINUE to indicate that message validation is
     *         incomplete, and that a preliminary response was returned as the
     *         response message in messageInfo.
     *         <p/>
     *         When this status value is returned to challenge an
     *         application request message, the challenged request must be saved
     *         by the authentication module such that it can be recovered
     *         when the module's validateRequest message is called to process
     *         the request returned for the challenge.
     *         <p/>
     *         <li> AuthStatus.SEND_FAILURE to indicate that message validation failed
     *         and that an appropriate failure response message is available by
     *         calling getResponseMessage on messageInfo.
     *         </ul>
     *         throws AuthException When the message processing failed without
     *         establishing a failure response message (in messageInfo).
     */
    public AuthStatus validateRequest(MessageInfo messageInfo,
                                      Subject clientSubject,
                                      Subject serviceSubject) throws AuthException {

        assert (messageInfo.getMap().containsKey(IS_MANDATORY_INFO_KEY) == mandatory);

        HttpServletRequest request = (HttpServletRequest) messageInfo.getRequestMessage();
        HttpServletResponse response = (HttpServletResponse) messageInfo.getResponseMessage();
        debugRequest(request);

        // should specify encoder
        String authorization = request.getHeader(AUTHORIZATION_HEADER);

        if (authorization != null && authorization.startsWith(NEGOTIATE)) {

            authorization = authorization.substring(NEGOTIATE.length() + 1);

            // should specify a decoder
            byte[] requestToken = Base64.decodeBase64(authorization.getBytes());
            GSSContext gssContext = null;

            try {

                gssContext = gssManager.createContext((GSSCredential) null);

                byte[] gssToken = gssContext.acceptSecContext(requestToken, 0, requestToken.length);

                if (gssToken != null) {
                    byte[] responseToken = Base64.encodeBase64(gssToken);
                    response.setHeader(AUTHENTICATION_HEADER, "Negotiate" + responseToken);
                    debugToken("jmac.servlet.authentication.token", responseToken);
                }

                if (!gssContext.isEstablished()) {

                    if (debug || LOG.isLoggable(Level.FINE)) {
                        LOG.log(debugLevel, "jmac.gss_dialog_continued");
                    }

                    response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
                    return AuthStatus.SEND_CONTINUE;

                } else {

                    String mechID;
                    try {
                        Oid oid = gssContext.getMech();
                        mechID = oid.toString();
                    } catch (GSSException gsse) {
                        mechID = "Undefined GSS Mechanism";

                        if (debug || LOG.isLoggable(Level.FINE)) {
                            LOG.log(debugLevel, "jmac.gss_mechanism_undefined", gsse);
                        }
                    }

                    GSSName name = gssContext.getSrcName();

                    if (!setCallerPrincipal(messageInfo, gssContext, clientSubject)) {
                        return sendFailureMessage(response, "Failed setting caller principal");
                    }

                    /* we may need to add something like a cookie to the
                  * response (that will be returned in subsequent requests).
                  * At this point, I am presuming that the browser will
                  * resend the authorization token.
                  */
                    messageInfo.getMap().put(AUTH_TYPE_INFO_KEY, mechID);

                    if (debug || LOG.isLoggable(Level.FINE)) {
                        LOG.log(debugLevel, "jmac.gss_dialog_complete");
                    }

                }

            } catch (GSSException gsse) {
                LOG.severe("GSSException occurred. Major Error: " + gsse.getMajorString()
                        + ". Minor Error: " + gsse.getMinorString());

                if (requestToken != null) {
                    debugToken("jmac.servlet.authorization.token", requestToken);
                    if (isNTLMToken(requestToken)) {
                        // until we add support for NTLM
                        return sendFailureMessage(response, "No support for NTLM");
                    }
                }

                if (debug || LOG.isLoggable(Level.FINE)) {
                    LOG.log(debugLevel, "jmac.gss_dialog_failed", gsse);
                }

                // for other errors throw an AuthException

                AuthException ae = new AuthException();
                ae.initCause(gsse);
                throw ae;
            } finally {
                if (gssContext != null) {
                    try {gssContext.dispose();} catch (GSSException e) {}
                }
            }

        } else if (mandatory) {

            response.setHeader(AUTHENTICATION_HEADER, NEGOTIATE);
            response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);

            if (debug || LOG.isLoggable(Level.FINE)) {
                LOG.log(debugLevel, "jmac.sevlet_header_added_to_response", NEGOTIATE);
            }
            return AuthStatus.SEND_CONTINUE;

        } else if (authorization != null) {
            LOG.warning("jmac.servlet_authorization_header_ignored");
        }

        return AuthStatus.SUCCESS;
    }


    /**
     * Secure a service response before sending it to the client.
     * <p/>
     * This method is called to transform the response message acquired by
     * calling getResponseMessage (on messageInfo) into the mechanism-specific
     * form to be sent by the runtime.
     * <p> This method conveys the outcome of its message processing either
     * by returning an AuthStatus value or by throwing an AuthException.
     *
     * @param messageInfo    A contextual object that encapsulates the
     *                       client request and server response objects, and that may be
     *                       used to save state across a sequence of calls made to the
     *                       methods of this interface for the purpose of completing a
     *                       secure message exchange.
     * @param serviceSubject A Subject that represents the source of the
     *                       service
     *                       response, or null. It may be used by the method implementation
     *                       to retrieve Principals and credentials necessary to secure
     *                       the response. If the Subject is not null,
     *                       the method implementation may add additional Principals or
     *                       credentials (pertaining to the source of the service
     *                       response) to the Subject.
     * @return An AuthStatus object representing the completion status of
     *         the processing performed by the method.
     *         The AuthStatus values that may be returned by this method
     *         are defined as follows:
     *         <p/>
     *         <ul>
     *         <li> AuthStatus.SEND_SUCCESS when the application response
     *         message was successfully secured. The secured response message may be
     *         obtained by calling getResponseMessage on messageInfo.
     *         <p/>
     *         <li> AuthStatus.SEND_CONTINUE to indicate that the application response
     *         message (within messageInfo) was replaced with a security message
     *         that should elicit a security-specific response (in the form of a
     *         request) from the peer.
     *         <p/>
     *         This status value serves to inform the calling runtime that
     *         (to successfully complete the message exchange) it will
     *         need to be capable of continuing the message dialog by processing
     *         at least one additional request/response exchange (after having
     *         sent the response message returned in messageInfo).
     *         <p/>
     *         When this status value is returned, the application response must
     *         be saved by the authentication module such that it can be recovered
     *         when the module's validateRequest message is called to process
     *         the elicited response.
     *         <p/>
     *         <li> AuthStatus.SEND_FAILURE to indicate that a failure occurred while
     *         securing the response message and that an appropriate failure response
     *         message is available by calling getResponseMeessage on messageInfo.
     *         </ul>
     *         throws AuthException When the message processing failed without
     *         establishing a failure response message (in messageInfo).
     */
    public AuthStatus secureResponse(MessageInfo messageInfo,
                                     Subject serviceSubject) throws AuthException {
        LOG.fine("secureResponse called");
        return AuthStatus.SEND_SUCCESS;
    }

    /**
     * Remove method specific principals and credentials from the subject.
     *
     * @param messageInfo a contextual object that encapsulates the
     *                    client request and server response objects, and that may be
     *                    used to save state across a sequence of calls made to the
     *                    methods of this interface for the purpose of completing a
     *                    secure message exchange.
     * @param subject     the Subject instance from which the Principals and
     *                    credentials are to be removed.
     *                    throws AuthException If an error occurs during the Subject
     *                    processing.
     */
    public void cleanSubject(MessageInfo messageInfo, Subject subject) throws AuthException {
        LOG.fine("cleanSubject called");
    }

    /**
     * Sends a failure message in the response
     *
     * @param response
     * @param message
     * @return
     */
    AuthStatus sendFailureMessage(HttpServletResponse response, String message) {

        try {
            response.setStatus(HttpServletResponse.SC_FORBIDDEN);
            response.sendError(HttpServletResponse.SC_FORBIDDEN, message);
        } catch (Throwable t) {
            // status code has been set, and proper AuthStatus will be returned
            LOG.log(Level.WARNING, "jmac.servlet_failed_sending_failure", t);
        } finally {
            return AuthStatus.SEND_FAILURE;
        }
    }

    protected boolean setCallerPrincipal(MessageInfo messageInfo, GSSContext gssContext, 
            Subject clientSubject) throws AuthException, GSSException {

        // create Subject with principals from name
        Subject s = GSSUtil.createSubject(gssContext.getSrcName(), null);

        Set principals = s.getPrincipals();

        if (principals.size() > 0) {

            clientSubject.getPrincipals().addAll(principals);

            // if more than one principal, caller selection is unpredictable
            Principal caller = (Principal) principals.iterator().next();

            CallerPrincipalCallback callerPrincipalCallback = new CallerPrincipalCallback(clientSubject, caller);


            
            GroupPrincipalCallback groupPrincipalCallback = new GroupPrincipalCallback(clientSubject,
                    (caller == null ? null : getGroupsForCaller(messageInfo, gssContext)));

            try {
                handler.handle(new Callback[]{callerPrincipalCallback, groupPrincipalCallback});
                if (debug || LOG.isLoggable(Level.FINE)) {
                    LOG.log(debugLevel, "jmac.caller_principal",
                            new Object[]{caller});
                }


                return true;
            } catch (Exception e) {
                // should not happen
                LOG.log(Level.WARNING, "jmac.failed_to_set_caller", e);
            }

        }
        return false;
    }


    /**
     * Provides a list of group names to apply to the calling principal.
     * <p/>
     * This method is called in Glassfish V2    Build 41 each time resource with a security
     * constraint is called. Accordinly, this method should be very performant. A new instance
     * of this class is created for each call, so slow set up of resources within this class should be avoided.
     * <p/>
     * @param caller the calling principal. The Kerberos user.
     * @return an array of group names.
     */
    public abstract String[] getGroupsForCaller(MessageInfo messageInfo, GSSContext gssContext 
            ) throws AuthException;


    /**
     * @param bytes
     * @return true if the token is an NTLM token
     */
    boolean isNTLMToken(byte[] bytes) {

        String s = new String(bytes);
        return s.startsWith(NTLM_INITIAL_TOKEN);
    }

    private void debugToken(String message, byte[] bytes) {

        if (debug || LOG.isLoggable(Level.FINE)) {

            StringBuffer sb = new StringBuffer();
            sb.append("\n");
            sb.append("Token " + (Base64.isArrayByteBase64(bytes) ? "is" : "is Not") + " Base64 encoded" + "\n");
            sb.append("bytes: ");
            boolean first = true;
            for (byte b : bytes) {
                int i = b;
                if (first) {
                    sb.append(i);
                    first = false;
                } else {
                    sb.append(", " + i);
                }
            }
            LOG.log(debugLevel, message, sb);
        }
    }

    private void debugRequest(HttpServletRequest request) {

        if (debug || LOG.isLoggable(Level.FINE)) {
            StringBuffer sb = new StringBuffer();
            sb.append("\n");
            try {
                sb.append("Request: " + request.getRequestURL() + "\n");
                sb.append("UserPrincipal: " + request.getUserPrincipal() + "\n");
                sb.append("AuthType: " + request.getAuthType() + "\n");
                sb.append("Headers:" + "\n");
                Enumeration names = request.getHeaderNames();
                while (names.hasMoreElements()) {
                    String name = (String) names.nextElement();
                    sb.append("\t" + name + "\t" + request.getHeader(name) + "\n");
                }
                LOG.log(debugLevel, "jmac.servlet_request", sb);

            } catch (Throwable t) {
                LOG.log(Level.WARNING, "jmac.servlet_debug_request", t);
            }
        }
    }


}
